Mouse mouse;
spork ~ mouse.start(0);
spork ~ mouse.selfUpdate();


// sequencer params
120 => int BPM; // beats per minute
(1.0/BPM) :: minute / 2.0 => dur STEP; // step duration
8 => int NUM_STEPS; //steps per sequence

1.0 => float div;



// scene setup
GG.scene() @=> GScene @ scene;
GG.camera() @=> GCamera @ cam;
GG.scene().backgroundColor(@(0.871, 0.722, 0.529));
cam.orthographic();


GGen drumGroupTop --> GG.scene();
GGen drumGroupBot --> GG.scene();

GPad drumPadsTop[NUM_STEPS];
GPad drumPadsBot[NUM_STEPS];


// update pad positions on window resize
fun void resizeListener() {
    placePads();
    WindowResizeEvent e;  // now listens to the window resize event
    while (true) {
        e => now;  // window has been resized!
        <<< GG.windowWidth(), " , ", GG.windowHeight() >>>;
        placePads();
    }
} spork ~ resizeListener();


(GG.frameWidth() * 1.0) / (GG.frameHeight() * 1.0) => float aspect;

cam.viewSize() => float frustrumHeight;
frustrumHeight * aspect => float frustrumWidth;
frustrumWidth / NUM_STEPS => float padSpacing;


fun void placePads(){

    placePadsHorizontal(
        drumPadsTop, drumGroupTop,
        frustrumWidth - 4,
        - frustrumHeight/2.0 + padSpacing + 0.3
    );

    placePadsHorizontal(
        drumPadsBot, drumGroupBot,
        frustrumWidth - 4,
        - frustrumHeight/2.0 + padSpacing/2.0
    );





}


// places along horizontal axis
fun void placePadsHorizontal(GPad pads[], GGen @ parent, float width, float y) {
    width / pads.size() => float padSpacing;
    for (0 => int i; i < pads.size(); i++) {
        pads[i] @=> GPad pad;

        // initialize pad
        pad.init(mouse);

        // connect to scene
        pad --> parent;

        // set transform
        pad.sca(padSpacing * .3);
        pad.posX(padSpacing * i - width / 2.0 + padSpacing / 2.0);
    }
    parent.posY(y);  // position the entire row
}


// places along vertical axis
fun void placePadsVertical(GPad pads[], GGen @ parent, float height, float x)
{
    // scale height down a smidge
    // .95 *=> height;
    height / pads.size() => float padSpacing;
    for (0 => int i; i < pads.size(); i++) {
        pads[i] @=> GPad pad;

        // initialize pad
        pad.init(mouse);

        // connect to scene
        pad --> parent;

        // set transform
        pad.sca(padSpacing * .7);
        pad.posY(padSpacing * i - height / 2.0 + padSpacing / 2.0);
    }
    parent.posX(x);  // position entire column
}

// base class for percussion instruments
class Instrument extends Chugraph {
    fun void play() {}
}



class Taiko1 extends Instrument { 
    inlet => Noise n => LPF f => ADSR e => outlet;
    200 => f.freq;
    40 => f.gain;
    e.set(5::ms, 50::ms, 0.1, 100::ms);

    fun void play() {
        e.keyOn();
        50::ms => now;
        e.keyOff();
        e.releaseTime() => now;
    }
}


class Taiko2 extends Instrument {  
    inlet => Noise n => LPF f => ADSR e => outlet;
    300 => f.freq;
    40 => f.gain;
    e.set(10::ms, 40::ms, 0.1, 80::ms);

    fun void play() {
        e.keyOn();
        50::ms => now;
        e.keyOff();
        e.releaseTime() => now;
    }

}


// SEQUENCER !!
Gain main => JCRev rev => dac;  // main bus
.1 => main.gain;
0.1 => rev.mix;


Taiko1 drum1 => main;
Taiko2 drum2 => main;

// randomize initial drum pattern
for (int i; i < NUM_STEPS; i++) {
    // 35% chance of kick
    if (Math.randomf() < .35) drumPadsTop[i].activate();
    // 45% chance of snare
    if (Math.randomf() < .45) drumPadsBot[i].activate();
}






// sequence percussion (monophonic)
fun void sequenceBeat(Instrument @ instrument, GPad pads[], int rev, dur step) {
    0 => int i;
    if (rev) pads.size() - 1 => i;


    while (true) {
        false => int juice;
        if (pads[i].active()) {
            true => juice;
            spork ~ instrument.play();  // play sound
        }
        // start animation
        pads[i].play(juice);  // must happen after .active() check
        // pass time
        step => now;
        // stop animation
        pads[i].stop();

        // bump index, wrap around playhead to other end
        if (rev) {
            i--;
            if (i < 0) pads.size() - 1 => i;
        } else {
            i++;
            if (i >= pads.size()) 0 => i;
        }
    }
} 


spork ~ sequenceBeat(drum1, drumPadsTop, false, STEP / 1.0);
spork ~ sequenceBeat(drum2, drumPadsBot, false, STEP / 2.0);



// BACKGROUNDS =========================

GPlane nintendo --> GG.scene();
FlatMaterial skin;

FileTexture switchSkin;
switchSkin.path(me.dir() + "switch.png");

skin.diffuseMap(switchSkin);
nintendo.mat(skin);
nintendo.scaX(frustrumWidth/1.05);
nintendo.scaY(frustrumHeight/3.5);
nintendo.translateY(-2.2);
nintendo.posZ(-0.5);


GPlane eurorack --> GG.scene();
FlatMaterial base;

FileTexture baseSkin;
baseSkin.path(me.dir() + "tv.png");

base.diffuseMap(baseSkin);
eurorack.mat(base);
eurorack.scaX(frustrumWidth);
eurorack.scaY(frustrumHeight/1.4);
eurorack.translateY(1.2);
eurorack.posZ(-1.2);


// ADD TAIKO
GMesh taiko;
NormalsMaterial mat;

taiko.set( // load the obj file and assign it to the mesh geometry
    ObjLoader.load("new.obj"),   // set obj geometry
    mat                                                       // set material
);

taiko --> GG.scene();
taiko.translateY(1);
taiko.sca(0.25);
taiko.rot(@(4.6, 0, 0));


GMesh bunny;


bunny.set( // load the obj file and assign it to the mesh geometry
    ObjLoader.load("bunny.obj"),   // set obj geometry
    mat                                                       // set material
);



bunny --> GG.scene();
bunny.posZ(-5);
bunny.translateY(0.5);
bunny.translateX(-2.5);
bunny.sca(0.1);
bunny.rot(@(4.6, 0, 0));

GMesh bunny2;


bunny2.set( // load the obj file and assign it to the mesh geometry
    ObjLoader.load("bunny.obj"),   // set obj geometry
    mat                                                       // set material
);



bunny2 --> GG.scene();
bunny2.posZ(-5);
bunny2.translateY(0.5);
bunny2.translateX(2.5);
bunny2.sca(0.1);
bunny2.rot(@(4.6, 0, 0));




0 @=> int x2;
0 @=> int y2;


// ROTATION MECHANISM ====================
Event beatEvent;

int rotationDirection;

fun void rotateOnBeat(){
    1 => rotationDirection;
    while(true){
        beatEvent => now;

        taiko.rotateX(rotationDirection*Math.PI/8);
        //taiko.posY(rotationDirection);
        bunny.rotateZ(rotationDirection*Math.PI/10);
        bunny2.rotateZ(-rotationDirection*Math.PI/10);
        -1 *=> rotationDirection;

    }
    
} spork ~rotateOnBeat();


fun void sequence(){
    while(true){
        STEP => now;
        beatEvent.broadcast();
    }
} spork ~sequence();



// DDR ======================================
GG.dt() => float dt;


fun void initArrow(){
    while(true){
            // test for the 'a' key
        if( KB.isKeyDown(KB.KEY_A) )
        {
            GCube leftBlock --> GG.scene();
            leftBlock.posY(-2.5);
            leftBlock.sca(@(0.8, 0.4, 0.1));

            FileTexture dancePad;
            dancePad.path("boom.png");
            leftBlock.mat().diffuseMap(dancePad);
            leftBlock.posX(-2);

            <<< "A key is down, now:", now >>>;


            while(leftBlock.posY() < 1.5){
                <<< leftBlock.posY(), now >>>;
                leftBlock.posY(leftBlock.posY()+0.05);
                // GG.nextFrame() => now;
                // 0.25::second => now;
                // float p;
                // 0.05 +=> p;
                // leftBlock.posY(p);
                
                
                
            } spork ~ dance(leftBlock);
            <<< leftBlock.posY() + "TEST", now >>>;
             
        }
        
        if(KB.isKeyDown(KB.KEY_D)){
            GCube rightBlock --> GG.scene();
            rightBlock.posY(-2.5);
            rightBlock.sca(@(0.8, 0.4, 0.1));

            FileTexture dancePad;
            dancePad.path("boom.png");
            rightBlock.mat().diffuseMap(dancePad);
            rightBlock.posX(2);

            <<< "A key is down, now:", now >>>;


            while(rightBlock.posY() < 1.5){
                <<< rightBlock.posY(), now >>>;
                rightBlock.posY(rightBlock.posY()+0.05);
                // float p;
                // 0.05 +=> p;
                // rightBlock.posY(p);
                
                // GG.nextFrame() => now;
                // 0.25::second => now;
                
                
                
            } spork ~ dance(rightBlock);
            <<< rightBlock.posY() + "TEST", now >>>;
             
                
        }

        if(KB.isKeyDown(KB.KEY_W)){
            bunny.posZ(0);
            spork ~ playAppegiatorUp();
            makeCirclesLeft();


                
        }

        if(KB.isKeyDown(KB.KEY_S)){
            bunny2.posZ(0);
            spork ~ playAppegiatorDown();
            makeCirclesRight();
            

        }
        30::ms => now;    

    }

} spork ~ initArrow();


class Circle extends GGen
{
    // for drawing our circle 
    GLines circle --> this;
    // randomize rate
    Math.random2f(2,2.5) => float rate;
    // default color
    color( @(	1.000, 0.714, 0.757) );

    // initialize a circle
    fun void init( int resolution, float radius )
    {
        // incremental angle from 0 to 2pi in N steps
        2*pi / resolution => float theta;    
        // positions of our circle
        vec3 pos[resolution];
        // previous, init to 1 zero
        @(radius,0) => vec3 prev;
        // loop over vertices
        for( int i; i < pos.size(); i++ )
        {
            // rotate our vector to plot a circle
            // https://en.wikipedia.org/wiki/Rotation_matrix
            Math.cos(theta)*prev.x - Math.sin(theta)*prev.y => pos[i].x;
            Math.sin(theta)*prev.x + Math.cos(theta)*prev.y => pos[i].y;
            // just XY here, 0 for Z
            0 => pos[i].z;
            // remember v as the new previous
            pos[i] => prev;
        }
        
        // set positions
        circle.geo().positions( pos );
    }
    
    fun void color( vec3 c )
    {
        circle.mat().color( c );
    }
    
    fun void update( float dt )
    {
        .005 + .015*Math.sin(now/second*rate) => float s;
        circle.sca( @(s,s,s ) );
    }
}

fun void makeCirclesLeft(){
    8 => int NUM_CIRCLES;
    10 => int N;
    3 => int RADIUS;
    // an array of our custom circles
    Circle circles[NUM_CIRCLES];
    // iterate over circles array
    for( auto circ : circles )
    {
        // initialize each 
        circ.init( N, RADIUS );
        // connect it
        circ --> GG.scene();
        // randomize location in XY
        @( -2.5,
            Math.random2f(0.5,2),
            1)  => circ.pos;
        // @( 0, 0, 1 ) => circ.position;

    }
    3::second => now;
    for( auto circ : circles )
    {
        circ --< GG.scene();
    }
}

fun void makeCirclesRight(){
    8 => int NUM_CIRCLES;
    10 => int N;
    3 => int RADIUS;
    // an array of our custom circles
    Circle circles[NUM_CIRCLES];
    // iterate over circles array
    for( auto circ : circles )
    {
        // initialize each 
        circ.init( N, RADIUS );
        // connect it
        circ --> GG.scene();
        // randomize location in XY
        @( 2.5,
            Math.random2f(0.5,2),
            1 )  => circ.pos;
        // @( 0, 0, 1 ) => circ.position;
    }
    3::second => now;
    for( auto circ : circles )
    {
        circ --< GG.scene();
    }
}









fun void dance(GCube block){
    GG.nextFrame() => now;
    block --< GG.scene();
    //insert sound code
   



    if(block.posX() == -2){
        taiko.rotateZ(-Math.PI/6); 
         <<<" DANCE TEST", now >>>;
         spork ~ playDrumLeft(); 
         70::ms => now;
         taiko.rotateZ(Math.PI/8); 

    } 

    if(block.posX() == 2){
        taiko.rotateZ(Math.PI/6); 
            
        spork ~ playDrumRight();
        70::ms => now;
        taiko.rotateZ(-Math.PI/8);  
    } 


    while (true) { 
        // rotate center cube


        //taiko.rotateOnLocalAxis(@(0.01, 0, 0), Math.atan(dt));
        //taiko.rotateOnLocalAxis(@(0, 0, 0.0001), Math.atan(dt));
        GG.nextFrame() => now; 
    }
    
}


fun void DrumSound(){
    Noise s => Gain g1 => LPF l => Gain out;
    Impulse i => Gain g2 => LPF l2 => g1;

    3 => g1.op;
    l.set(3000, 4);
    g2.gain(1.0 - 1.0/3000);
    g1.gain(200);

    while( true ) {
        0.8 => i.next;
        100::ms => now;
    }

    out => dac;

}

fun void playDrumLeft() {
    // note: connect output to external sources to connect
    Noise s => Gain s_env => LPF s_f => Gain output; // white noise source
    Impulse i => Gain g => Gain g_fb => g => LPF g_f => s_env;
   
    3 => s_env.op; // make s envelope a multiplier
    s_f.set(3000, 4); // set default drum filter
    g_fb.gain(1.0 - 1.0/800); // set default drum decay
    g_f.set(200, 1); // set default drum attack


    output => dac;
    0.6 => i.next;
    2::second => now;
    // g_fb.gain(1.0 - 1.0 / 10000);
    // s_f.set(5000, 5);
    // 0.8 => i.next;
    // 2::second => now;
}

fun void playDrumRight() {
    // note: connect output to external sources to connect
    Noise s => Gain s_env => LPF s_f => Gain output; // white noise source
    Impulse i => Gain g => Gain g_fb => g => LPF g_f => s_env;
   
    3 => s_env.op; // make s envelope a multiplier
    g_f.set(200, 1); // set default drum attack

    output => dac;
    g_fb.gain(1.0 - 1.0 / 1000);
    s_f.set(5000, 5);
    0.6 => i.next;
    2::second => now;
}


fun void playAppegiatorUp(){
    440 => float baseFrequency; // starting frequency
    5 => int numHarmonics; // number of harmonics to play
    250::ms => dur noteDur; // note duration

    SinOsc osc => dac;
    osc.gain(0.5);

    // for (0 => int x; x< 2; x++)
    // {
        // Loop through the number of harmonics
        for (0 => int i; i < numHarmonics; i++)
        {
            // Update the oscillator frequency to the next harmonic
            baseFrequency + (i * baseFrequency)  => osc.freq;
            // Advance time to play the note
            noteDur => now;
            baseFrequency - (i * baseFrequency)  => osc.freq;
            noteDur => now;
        }
    // }
}

fun void playAppegiatorDown(){
    240 => float baseFrequency; // starting frequency
    3 => int numHarmonics; // number of harmonics to play
    250::ms => dur noteDur; // note duration

    SinOsc osc => dac;
    osc.gain(0.5);

    // for (0 => int x; x< 2; x++)
    // {
        // Loop through the number of harmonics
        for (5 => int i; i > 0; i--)
        {
            // Update the oscillator frequency to the next harmonic
            baseFrequency - (i * baseFrequency)  => osc.freq;
            // Advance time to play the note
            noteDur => now;
            baseFrequency + (i * baseFrequency)  => osc.freq;
            noteDur => now;
        }
    // }
}


// UI =========================================
// UI_Window window;


// UI_SliderFloat islider;
// fslider.range(60, 120);

// 120 => int BPM2;


// window.add(islider);
// islider.val(BPM2);
// window.text("[REMOTE]");




// fun void ISliderListener(UI_SliderInt @ islider){
//     while (true) {
//         islider.val() => BPM2;
//         islider => now;
//     }
// } spork ~ ISliderListener(islider);


// Game loop =========================================================
while (true) { 
    // rotate center cube
    GG.dt() => float dt;
    // taiko.sca(0.25+(Math.atan(dt)));


    //taiko.rotateOnLocalAxis(@(0.01, 0, 0), Math.atan(dt));
    //taiko.rotateOnLocalAxis(@(0, 0, 0.0001), Math.atan(dt));
    GG.nextFrame() => now; 
}